const Error = require('@cardstack/plugin-utils/error');
const { Pool } = require('pg');
const { range }  = require('lodash');
const rowToDocument = require('./row-to-doc');
const NameMapper = require('./name-mapper');

const safeIdentifier = /^[a-zA-Z0-9._]+$/;

module.exports = class Writer {
  static create(...args) {
    return new this(...args);
  }
  constructor({ config, dataSource, renameColumns, renameTables }) {
    this.config = config;
    this.pool = null;
    this.schema = null;
    this.dataSource = dataSource;
    this.mapper = new NameMapper(renameTables, renameColumns);
  }

  _getPool() {
    if (!this.pool) {
      let { user, host, database, password, port } = this.config;
      this.pool = new Pool({
        user,
        host,
        database,
        password,
        port
      });
    }
    return this.pool;
  }

  async _getSchema() {
    if (!this.schema) {
      let updater = await this.dataSource.indexer.beginUpdate();
      try {
        this.schema = await updater.schema();
      } finally {
        await updater.destroy();
      }
    }
    return this.schema;
  }

  async teardown() {
    await this.pool.end();
  }

  async prepareCreate(session, type, document, isSchema) {
    let { schema, table, args, columns } = this._prepareQuery(isSchema, type, document);
    if (document.id != null) {
      args.push(document.id);
      columns.push('id');
    }
    let placeholders = range(1, args.length + 1).map(n => `$${n}`).join(',');
    let client = await this._getPool().connect();
    try {
      await client.query('begin');
      let result = await client.query(`insert into ${schema}.${table} (${columns.join(',')}) values (${placeholders}) returning *`, args);
      let finalDocument = rowToDocument(this.mapper, await this._getSchema(), type, result.rows[0]);
      return {
        finalDocument,
        finalizer: finalize,
        aborter: abort,
        client
      };
    } catch (err) {
      await client.release();
      rethrowNicerError(err);
    }
  }

  async prepareUpdate(session, type, id, document, isSchema) {
    let { schema: dbschema, table, args, columns } = this._prepareQuery(isSchema, type, document);
    let client = await this._getPool().connect();
    let schema = await this._getSchema();
    args.push(id);
    try {
      await client.query('begin');
      let result = await client.query(`select * from ${dbschema}.${table} where id=$1`, [ id ]);
      if (result.rows < 1) {
        throw new Error("Not found", { status: 404, source: { pointer: '/data/id' } });
      }
      let initialDocument = rowToDocument(this.mapper, schema, type, result.rows[0]);
      result = await client.query(`update ${dbschema}.${table} set ${columns.map((name,index) => `${name}=$${index+1}`).join(',')} where id=$${args.length} returning *`, args);
      let finalDocument = rowToDocument(this.mapper, schema, type, result.rows[0]);
      return {
        originalDocument: initialDocument,
        finalDocument,
        finalizer: finalize,
        aborter: abort,
        client
      };
    } catch (err) {
      await client.release();
      rethrowNicerError(err);
    }
  }

  async prepareDelete(session, version, type, id, isSchema) {
    let { schema, table } = this._prepareQuery(isSchema, type);
    let client = await this._getPool().connect();
    try {
      await client.query('begin');
      let initialDocument = rowToDocument(this.mapper, await this._getSchema(), type, await client.query(`select * from ${schema}.${table} where id=$1`, [ id ]));
      await client.query(`delete from ${schema}.${table} where id=$1`, [id]);
      return {
        originalDocument: initialDocument,
        finalizer: finalize,
        aborter: abort,
        client
      };
    } catch (err) {
      await client.release();
      rethrowNicerError(err);
    }

  }

  _prepareQuery(isSchema, type, document) {
    if (isSchema) {
      throw new Error("The @cardstack/postgresql data source does not support schema creation", { status: 400 });
    }
    if (!this.config) {
      throw new Error(`No db configuration`);
    }

    let { schema, table } = this.mapper.tableForType(type);
    if (!safeIdentifier.test(table)) {
      throw new Error(`Disallowed table name ${table}`);
    }
    if (!safeIdentifier.test(schema)) {
      throw new Error(`Disallowed schema name ${schema}`);
    }

    let args = [];
    let columns = (document ? Object.entries(document.attributes || {}) : []).map(([key, value]) => {
      key = this.mapper.columnNameFor(schema, table, key);
      if (!safeIdentifier.test(key)) {
        throw new Error(`Disallowed column name ${key}`);
      }
      if (value instanceof Array) {
        // Arrays are a valid type of json to store in a json column, but they
        // are not escaped well when used with the prepared statement
        value = JSON.stringify(value);
      }
      args.push(value);
      return quoteKey(key);
    });

    // Handle relationships - TODO: no handling of polymorphism here yet
    (document ? Object.entries(document.relationships || {}) : []).map(([key, value]) => {
      key = this.mapper.columnNameFor(schema, table, key);
      if (value && value.data && value.data.id) {
        columns.push(quoteKey(key));
        args.push(value.data.id);
      }
    });
    return { table, schema, args, columns };
  }
};

function quoteKey(key) {
  return `"${key}"`;
}

async function finalize(pendingChange) {
  let { client } = pendingChange;
  await client.query('commit');
  await client.release();
}

async function abort(pendingChange) {
  let { client } = pendingChange;
  await client.query('rollback');
  await client.release();
}

function rethrowNicerError(err) {
  if (err.constraint) {
    throw new Error(err.message, { title: "Constraint violation", status: 401 });
  }
  throw err;
}
